/* gzin.c -- Input of .gz files
 *  -- Howard Chu, November 19, 2001
 * Copyright (C) 2001-2002, Howard Chu
 *
 * Stripped down from gzio.c by Jean-Loup Gailly.
 * Copyright (C) 1995-1998 Jean-loup Gailly.
 * For conditions of distribution and use, see copyright notice in zlib.h
 *
 */

/* @(#) $Id: gzin.c,v 1.2 2002/02/03 02:23:41 hyc Exp $ */

#include <stdio.h>

#include "zutil.h"

struct internal_state {int dummy;}; /* for buggy compilers */

#ifndef Z_BUFSIZE
#  ifdef MAXSEG_64K
#    define Z_BUFSIZE 4096 /* minimize memory usage for 16-bit DOS */
#  else
#    define Z_BUFSIZE 16384
#  endif
#endif
#ifndef Z_PRINTF_BUFSIZE
#  define Z_PRINTF_BUFSIZE 4096
#endif

#define TRYFREE(p) {if (p) free(p);}

static int gz_magic[2] = {0x1f, 0x8b}; /* gzip magic header */

/* gzip flag byte */
#define ASCII_FLAG   0x01 /* bit 0 set: file probably ascii text */
#define HEAD_CRC     0x02 /* bit 1 set: header CRC present */
#define EXTRA_FIELD  0x04 /* bit 2 set: extra field present */
#define ORIG_NAME    0x08 /* bit 3 set: original file name present */
#define COMMENT      0x10 /* bit 4 set: file comment present */
#define RESERVED     0xE0 /* bits 5..7: reserved */

typedef struct gz_stream {
    z_stream stream;
    int      z_err;   /* error code for last stream operation */
    Byte     *outbuf;  /* output buffer */
    FILE     *fp;	/* decompressed output stream */
    uLong    crc;     /* crc32 of uncompressed data */
#define	Z_INIT	1
#define	Z_EOF	2
#define	Z_TRANS	4
    int      flags; /* see above */
} gz_stream;

local int    get_byte     OF((gz_stream *s));
local void   check_header OF((gz_stream *s));
local int    destroy      OF((gz_stream *s));
local uLong  getLong      OF((gz_stream *s));

/* ===========================================================================
     Opens a gzip (.gz) file for reading.
     gz_open return NULL if the file could not be opened or if there was
   insufficient memory to allocate the (de)compression state; errno
   can be checked to distinguish the two cases (if errno is zero, the
   zlib error is Z_MEM_ERROR).
*/
gzFile gz_input (FILE *out, gzFile *f)
{
    int err;
    gz_stream *s;

    if (f)
    {
    	s = (gz_stream *)f;
	inflateEnd(&s->stream);
    }
    else
    {
    	s = (gz_stream *)malloc(sizeof(gz_stream) + Z_BUFSIZE);
	if (!s) return Z_NULL;
	s->outbuf = (Byte*)(s+1);
    }
    s->crc = crc32(0L, Z_NULL, 0);
    s->fp = out;
    s->z_err = 0;
    s->flags = 0;
    memset(&s->stream, 0, sizeof(s->stream));

    err = inflateInit2(&(s->stream), -MAX_WBITS);
    /* windowBits is passed < 0 to tell that there is no zlib header.
    * Note that in this case inflate *requires* an extra "dummy" byte
    * after the compressed stream in order to complete decompression and
    * return Z_STREAM_END. Here the gzip CRC32 ensures that 4 bytes are
    * present after the compressed stream.
    */
    if (err != Z_OK) {
            return destroy(s), (gzFile)Z_NULL;
    }
    return (gzFile)s;
}

/* ===========================================================================
     Read a byte from a gz_stream; update next_in and avail_in. Return EOF
   for end of file.
   IN assertion: the stream s has been sucessfully opened for reading.
*/
local int get_byte(s)
    gz_stream *s;
{
    if (s->flags & Z_EOF) return EOF;
    if (s->stream.avail_in == 0)
    {
	s->flags |= Z_EOF;
	return EOF;
    }
    s->stream.avail_in--;
    return *(s->stream.next_in)++;
}

/* ===========================================================================
      Check the gzip header of a gz_stream opened for reading. Set the stream
    mode to transparent if the gzip magic header is not present; set s->err
    to Z_DATA_ERROR if the magic header is present but the rest of the header
    is incorrect.
    IN assertion: the stream s has already been created sucessfully;
       s->stream.avail_in is zero for the first time, but may be non-zero
       for concatenated .gz files.
*/
local void check_header(s)
    gz_stream *s;
{
    int method; /* method byte */
    int flags;  /* flags byte */
    uInt len;
    int c;

    /* Check the gzip magic header */
    for (len = 0; len < 2; len++) {
	c = get_byte(s);
	if (c != gz_magic[len]) {
	    if (len != 0) s->stream.avail_in++, s->stream.next_in--;
	    if (c != EOF) {
		s->stream.avail_in++, s->stream.next_in--;
		s->flags |= Z_TRANS;
	    }
	    s->z_err = s->stream.avail_in != 0 ? Z_OK : Z_STREAM_END;
	    return;
	}
    }
    method = get_byte(s);
    flags = get_byte(s);
    if (method != Z_DEFLATED || (flags & RESERVED) != 0) {
	s->z_err = Z_DATA_ERROR;
	return;
    }

    /* Discard time, xflags and OS code: */
    for (len = 0; len < 6; len++) (void)get_byte(s);

    if ((flags & EXTRA_FIELD) != 0) { /* skip the extra field */
	len  =  (uInt)get_byte(s);
	len += ((uInt)get_byte(s))<<8;
	/* len is garbage if EOF but the loop below will quit anyway */
	while (len-- != 0 && get_byte(s) != EOF) ;
    }
    if ((flags & ORIG_NAME) != 0) { /* skip the original file name */
	while ((c = get_byte(s)) != 0 && c != EOF) ;
    }
    if ((flags & COMMENT) != 0) {   /* skip the .gz file comment */
	while ((c = get_byte(s)) != 0 && c != EOF) ;
    }
    if ((flags & HEAD_CRC) != 0) {  /* skip the header crc */
	for (len = 0; len < 2; len++) (void)get_byte(s);
    }
    s->z_err = (s->flags & Z_EOF) ? Z_DATA_ERROR : Z_OK;
}

 /* ===========================================================================
 * Cleanup then free the given gz_stream. Return a zlib error code.
   Try freeing in the reverse order of allocations.
 */
local int destroy (s)
    gz_stream *s;
{
    int err = Z_OK;

    if (!s) return Z_STREAM_ERROR;

    if (s->stream.state != NULL) {
	err = inflateEnd(&(s->stream));
    }
    if (s->z_err < 0) err = s->z_err;

    TRYFREE(s);
    return err;
}

/* ===========================================================================
     Reads the given number of uncompressed bytes from the compressed file.
   gzread returns the number of bytes actually read (0 for end of file).
*/
long ZEXPORT gz_flush (file, buf, len)
    gzFile file;
    voidp buf;
    unsigned len;
{
    gz_stream *s = (gz_stream*)file;
    long total = 0;

    if (s == NULL) return Z_STREAM_ERROR;
    if (s->z_err == Z_DATA_ERROR || s->z_err == Z_ERRNO) return -1;
    if (s->z_err == Z_STREAM_END) return 0;  /* EOF */

    s->stream.next_in = buf;
    s->stream.avail_in = len;

    if (!(s->flags & Z_INIT))
    {
	s->flags |= Z_INIT;
	check_header(s); /* skip the .gz header */
    }

    if (s->flags & Z_TRANS)
	return fwrite(buf, len, 1, s->fp);

    while (s->stream.avail_in)
    {
	s->stream.next_out = s->outbuf;
	s->stream.avail_out = Z_BUFSIZE;
        s->z_err = inflate(&(s->stream), Z_NO_FLUSH);
	len = s->stream.next_out - s->outbuf;
	s->crc = crc32(s->crc, s->outbuf, len);
	if (s->z_err == Z_STREAM_END)
	{
	    /* Check CRC and original size */
	    if (getLong(s) != s->crc)
		s->z_err = Z_DATA_ERROR;
	}
	total += len;
	if (fwrite(s->outbuf, len, 1, s->fp) < 1 ) break;
	if (s->z_err != Z_OK) break;
    }
    return total;
}

/* ===========================================================================
   Reads a long in LSB order from the given gz_stream. Sets z_err in case
   of error.
*/
local uLong getLong (s)
    gz_stream *s;
{
    uLong x = (uLong)get_byte(s);
    int c;

    x += ((uLong)get_byte(s))<<8;
    x += ((uLong)get_byte(s))<<16;
    c = get_byte(s);
    if (c == EOF) s->z_err = Z_DATA_ERROR;
    x += ((uLong)c)<<24;
    return x;
}
